import pytest

import nir
import numpy as np
import torch

from nirtorch.torch_tracer import torch_to_nir


def _filter_edges(graph, t1, t2):
    return [
        e
        for e in graph.edges
        if graph.nodes[e[0]].__class__ == t1 and graph.nodes[e[1]].__class__ == t2
    ]


def _filter_nodes(graph, t):
    return {k: v for k, v in graph.nodes.items() if v.__class__ == t}


def test_trace_unknown_leaf():
    class MyModule(torch.nn.Module):
        def forward(self, x):
            return x

    model = MyModule()
    with pytest.raises(ValueError):
        torch_to_nir(model, {})


def test_trace_default_linear():
    model = torch.nn.Linear(1, 1)
    graph = torch_to_nir(model, {})
    assert graph.__class__ == nir.Affine


def test_trace_mapped_module():
    class MyModule(torch.nn.Module):
        def forward(self, x):
            return x

    def map_my_module(module):
        return nir.Linear(np.array([[1]]))

    model = MyModule()
    graph = torch_to_nir(model, {MyModule: map_my_module})
    assert graph.__class__ == nir.Linear


def test_trace_mapped_module_stateless():
    class MyModule(torch.nn.Module):
        def __init__(self):
            super().__init__()
            self.lin = torch.nn.Linear(1, 1)

        def forward(self, x, state):
            return self.lin(x) + state

    model = MyModule()
    graph = torch_to_nir(model, {})
    assert len(graph.nodes) == 3
    assert len(graph.edges) == 2
    assert len(_filter_edges(graph, nir.Input, nir.Affine)) == 1
    assert len(_filter_edges(graph, nir.Affine, nir.Output)) == 1


def test_trace_mapped_module_stateful():
    class MyModule(torch.nn.Module):
        def __init__(self):
            super().__init__()
            self.lin = torch.nn.Linear(1, 1)
            self.state = torch.tensor([1.0])

        def forward(self, x):
            return self.lin(x) + self.state

    model = MyModule()
    graph = torch_to_nir(model, {})
    assert len(graph.nodes) == 3
    assert len(graph.edges) == 2
    assert len(_filter_edges(graph, nir.Input, nir.Affine)) == 1
    assert len(_filter_edges(graph, nir.Affine, nir.Output)) == 1


def test_trace_addition():
    class MyModule(torch.nn.Module):
        def __init__(self):
            super().__init__()
            self.lin = torch.nn.Linear(1, 1)

        def forward(self, x):
            return self.lin(x) + self.lin(x)

    model = MyModule()
    graph = torch_to_nir(model, {})
    assert len(graph.nodes) == 4
    assert len(graph.edges) == 4
    assert len(_filter_edges(graph, nir.Input, nir.Affine)) == 2
    assert len(_filter_edges(graph, nir.Affine, nir.Output)) == 2


def test_trace_sequential():
    model = torch.nn.Sequential(torch.nn.Linear(2, 1), torch.nn.Linear(1, 1))
    graph = torch_to_nir(model, {})
    assert graph.__class__ == nir.NIRGraph
    assert len(graph.nodes) == 4
    assert len(graph.edges) == 3
    list(graph.nodes.items())
    assert len(_filter_nodes(graph, nir.Input)) == 1, "We require one input node"
    assert len(_filter_nodes(graph, nir.Output)) == 1, "We require one output node"
    assert len(_filter_nodes(graph, nir.Affine)) == 2, "We require two affine nodes"

    assert len(_filter_edges(graph, nir.Input, nir.Affine)) == 1
    assert len(_filter_edges(graph, nir.Affine, nir.Affine)) == 1
    assert len(_filter_edges(graph, nir.Affine, nir.Output)) == 1


def test_trace_submodule():
    class MyModule(torch.nn.Module):
        def __init__(self):
            super(MyModule, self).__init__()
            self.linear = torch.nn.Linear(1, 1)

        def forward(self, x):
            return self.linear(x)

    model = MyModule()
    graph = torch_to_nir(model, {})
    assert graph.__class__ == nir.NIRGraph
    assert len(graph.nodes) == 3
    assert len(graph.edges) == 2
    list(graph.nodes.items())
    assert len(_filter_nodes(graph, nir.Input)) == 1, "We require one input node"
    assert len(_filter_nodes(graph, nir.Output)) == 1, "We require one output node"
    assert len(_filter_nodes(graph, nir.Affine)) == 1, "We require one affine node"

    assert len(_filter_edges(graph, nir.Input, nir.Affine)) == 1
    assert len(_filter_edges(graph, nir.Affine, nir.Output)) == 1


def test_trace_nested_submodule():
    class MyModule(torch.nn.Module):
        def __init__(self):
            super(MyModule, self).__init__()
            self.linear = torch.nn.Linear(1, 1)

        def forward(self, x):
            return self.linear(x)

    class MyModule2(torch.nn.Module):
        def __init__(self):
            super(MyModule2, self).__init__()
            self.module = MyModule()
            self.linear = torch.nn.Linear(1, 1)

        def forward(self, x):
            return self.linear(self.module(x))

    model = MyModule2()
    graph = torch_to_nir(model, {})
    assert graph.__class__ == nir.NIRGraph
    assert len(graph.nodes) == 4
    assert len(graph.edges) == 3

    assert len(_filter_nodes(graph, nir.Input)) == 1, "We require one input node"
    assert len(_filter_nodes(graph, nir.Output)) == 1, "We require one output node"
    assert len(_filter_nodes(graph, nir.Affine)) == 2, "We require two affine nodes"

    assert len(_filter_edges(graph, nir.Input, nir.Affine)) == 1
    assert len(_filter_edges(graph, nir.Affine, nir.Affine)) == 1
    assert len(_filter_edges(graph, nir.Affine, nir.Output)) == 1


def test_recursive_stateful():
    class MyModule(torch.nn.Module):
        def __init__(self):
            super().__init__()
            self.linear = torch.nn.Linear(1, 1)
            self.state = torch.tensor([1.0])

        def forward(self, x):
            out = self.linear(x + self.state)
            self.state = out
            return out

    model = MyModule()
    graph = torch_to_nir(model, {})
    assert graph.__class__ == nir.NIRGraph



import warnings
from typing import Callable, Optional, Union

import nir
import torch
import torch.nn as nn

from .graph import TorchGraph
from .graph_executor import GraphExecutor
from .utils import sanitize_name


def _switch_default_models(nir_graph: nir.NIRNode) -> Optional[torch.nn.Module]:
    node = None
    if isinstance(nir_graph, nir.Input):
        node = torch.nn.Identity()
        node.node_type = "input"
    elif isinstance(nir_graph, nir.Output):
        node = torch.nn.Identity()
        node.node_type = "output"
    return node


def _map_graph_to_torch(
    nir_graph: nir.NIRNode,
    model_map: Callable[[nn.Module], nn.Module],
    return_state: bool = True,
) -> nir.NIRNode:

    nodes = {}
    for name, node in nir_graph.nodes.items():
        mapped_module = model_map(node)
        if mapped_module is None:
            if isinstance(node, nir.NIRGraph):  # Recurse
                mapped_module = _map_graph_to_torch(node, model_map, return_state)
            else:  # Map identity nodes (input, output, etc.)
                mapped_module = _switch_default_models(node)
        nodes[sanitize_name(name)] = mapped_module
    # Ensure that we use legal PyTorch names for the edges
    sanitized_edges = [
        (sanitize_name(src), sanitize_name(dst)) for src, dst in nir_graph.edges
    ]
    # Reconstruct NIR graph with torch modules
    recon_graph = nir.NIRGraph(nodes, sanitized_edges)
    # Build a TorchGraph for tracing and executing
    trace_graph = TorchGraph.from_torch_modules(recon_graph.nodes, recon_graph.edges)
    # Build and return a graph executor module
    return GraphExecutor(trace_graph, return_state=return_state)


def load(
    nir_graph: Union[nir.NIRNode, str],
    model_map: Callable[[nir.NIRNode], nn.Module],
    return_state: bool = True,
) -> nn.Module:
    """
    DEPRECATED: Use `nirtorch.torch_to_nir` instead.

    Load a NIR graph and convert it to a torch module using the given model map.

    If you do not wish to operate with state, set `return_state=False`.

    Args:
        nir_graph (Union[nir.NIRNode, str]): The NIR object to load, or a string
            representing the path to the NIR object.
        model_map (Callable[[nn.NIRNode], nn.Module]): A method that returns the a torch
            module that corresponds to each NIR node.
        return_state (bool): If True, the execution of the loaded graph will return a
            tuple of [output, state], where state is a GraphExecutorState object.
            If False, only the NIR graph output will be returned. Note that state is
            required for recurrence to work in the graphs.

    Returns:
        nn.Module: The generated torch module
    """
    warnings.warn(
        "nirtorch.load is being deprecated in favour of nirtorch.nir_to_torch. "
        "Please refer to https://neuroir.org/docs/dev_pytorch.html for detailed instructions",
        DeprecationWarning,
    )

    if isinstance(nir_graph, str):
        nir_graph = nir.read(nir_graph)
    return _map_graph_to_torch(nir_graph, model_map, return_state=return_state)



from importlib.metadata import version as metadata_version, PackageNotFoundError

try:
    __version__ = version = metadata_version("nirtorch")
    del metadata_version
except PackageNotFoundError:
    # package is not installed
    pass

from .from_nir import load  # noqa F401
from .to_nir import extract_nir_graph  # noqa F401
from .nir_interpreter import nir_to_torch
from .torch_tracer import torch_to_nir

__all__ = [
    "extract_torch_graph",
    "extract_nir_graph",
    "load",
    "nir_to_torch",
    "torch_to_nir",
]



import typing
import warnings

import nir
import numpy as np
import torch
import pytest

from nirtorch import nir_interpreter


class LeakyNode(torch.nn.Module):
    def __init__(self, **kwargs):
        super().__init__()
        for key, value in kwargs.items():
            setattr(self, key, value)

    def forward(self, x, state=None):
        if state is None:
            state = self.v_leak
        new_value = state + self.tau * (self.r * x - self.v_leak)
        return new_value, new_value


def _map_li(li: nir.LI) -> torch.nn.Module:
    return LeakyNode(
        tau=torch.from_numpy(li.tau),
        r=torch.from_numpy(li.r),
        v_leak=torch.from_numpy(li.v_leak),
    )


def _map_affine_node(m: nir.Affine):
    lin = torch.nn.Linear(*m.weight.shape[-2:])
    lin.weight.data = torch.nn.Parameter(torch.tensor(m.weight).float())
    lin.bias.data = torch.nn.Parameter(torch.tensor(m.bias).float())
    return lin


def _map_linear_node(m: nir.Affine):
    lin = torch.nn.Linear(*m.weight.shape[-2:], bias=False)
    lin.weight.data = torch.nn.Parameter(torch.tensor(m.weight).float())
    return lin


def _map_identity(m: nir.NIRNode):
    return torch.nn.Identity()


_torch_node_map = {
    nir.Affine: _map_affine_node,
    nir.Linear: _map_linear_node,
    nir.CubaLIF: _map_identity,
    nir.LIF: _map_identity,
    nir.LI: _map_li,
}


def test_map_fails_on_unknown():
    w = np.ones((2, 2))
    linear = nir.Linear(w)
    graph = nir.NIRGraph.from_list(linear)
    with pytest.raises(ValueError):
        nir_interpreter.nir_to_torch(graph, {}, {})


def test_map_linear_node():
    w = np.random.random((2, 3)).astype(np.float32)
    linear = nir.Linear(w)
    module = nir_interpreter._map_nir_node_to_torch(linear, nir_interpreter.DEFAULT_MAP)
    assert torch.allclose(module.weight, torch.from_numpy(w))
    out = module(torch.ones(3))
    assert out.shape == (2,)


def test_map_conv1d_node():
    w = np.random.random((2, 3, 4)).astype(np.float32)
    b = np.random.random((2,)).astype(np.float32)
    nir_conv = nir.Conv1d(
        input_shape=10, weight=w, stride=2, padding=3, dilation=4, groups=1, bias=b
    )
    torch_conv = nir_interpreter._map_nir_node_to_torch(
        nir_conv, nir_interpreter.DEFAULT_MAP
    )
    assert torch.allclose(torch.from_numpy(nir_conv.weight), torch_conv.weight)
    assert torch.allclose(torch.from_numpy(nir_conv.bias), torch_conv.bias)
    assert nir_conv.stride == torch_conv.stride[0]
    assert nir_conv.padding == torch_conv.padding[0]
    assert nir_conv.dilation == torch_conv.dilation[0]
    assert nir_conv.groups == torch_conv.groups
    assert torch_conv(torch.ones(1, 3, 10)).shape == (1, 2, 2)


def test_map_conv2d_node():
    w = np.random.random((2, 3, 4, 5)).astype(np.float32)
    b = np.random.random((2,)).astype(np.float32)
    nir_conv = nir.Conv2d(
        input_shape=(10, 11),
        weight=w,
        stride=(2, 3),
        padding=(1, 2),
        dilation=(1, 2),
        groups=1,
        bias=b,
    )
    torch_conv = nir_interpreter._map_nir_node_to_torch(
        nir_conv, nir_interpreter.DEFAULT_MAP
    )
    assert torch.allclose(torch.from_numpy(nir_conv.weight), torch_conv.weight)
    assert torch.allclose(torch.from_numpy(nir_conv.bias), torch_conv.bias)
    assert nir_conv.stride == torch_conv.stride
    assert nir_conv.padding == torch_conv.padding
    assert nir_conv.dilation == torch_conv.dilation
    assert nir_conv.groups == torch_conv.groups
    assert torch_conv(torch.ones(1, 3, 10, 11)).shape == (1, 2, 5, 3)


def test_map_if_node():
    # Generating code with the "if" keyword can be sensitive
    # This tests ensures that it works properly
    v_th = np.random.random(1)
    r = np.random.random(1)
    nir_node = nir.IF(r=r, v_threshold=v_th)

    class MyIF(torch.nn.Module):
        def __init__(self, r, v_th):
            super().__init__()
            self.r = r
            self.v_th = v_th

        def forward(self, x, state):
            return x * self.r - self.v_th, state

    node_map = {
        nir.IF: lambda node: MyIF(
            torch.from_numpy(node.r), torch.from_numpy(node.v_threshold)
        )
    }
    torch_if = nir_interpreter._map_nir_node_to_torch(nir_node, node_map)
    data = torch.rand(1)
    assert torch.allclose(data * r - v_th, torch_if(data, None)[0])

    torch_if = nir_interpreter.nir_to_torch(nir.NIRGraph.from_list(nir_node), node_map)
    assert isinstance(torch_if.get_submodule("nir_node_if"), MyIF)


def test_map_leaky_stateful_graph_single_module():
    # Test that the graph can handle a single stateful module
    tau = np.random.random(1)
    r = np.random.random(1)
    v_leak = np.random.random(1)
    li = nir.LI(tau, r, v_leak)
    li_module = _map_li(li)
    module = nir_interpreter.nir_to_torch(nir.NIRGraph.from_list(li), {nir.LI: _map_li})
    data = torch.rand(1)
    output = module(data)
    assert isinstance(output, typing.Tuple)
    module_output = li_module(data)
    assert torch.allclose(output[0], module_output[0])
    assert torch.allclose(output[1]["li"], module_output[1])
    # Application a second time should yield a different, stateful response
    output = module(output[0], output[1])
    assert isinstance(output, typing.Tuple)
    module_output = li_module(*module_output)
    assert torch.allclose(output[0], module_output[0])


def test_map_leaky_stateful_graph_sequential_modules():
    # Test that the graph can handle two chained stateful modules internally
    tau = np.random.random(1)
    r = np.random.random(1)
    v_leak = np.random.random(1)
    li = nir.LI(tau, r, v_leak)
    li_module = _map_li(li)
    module = nir_interpreter.nir_to_torch(
        nir.NIRGraph.from_list(li, li), {nir.LI: _map_li}
    )
    data = torch.rand(1)
    output = module(data)
    assert isinstance(output, typing.Tuple)
    module_result = li_module(li_module(data)[0])[0]
    assert torch.allclose(output[0], module_result)


def test_map_linear_graph_default():
    w = np.random.random((2, 3)).astype(np.float32)
    linear = nir.Linear(w)
    graph = nir.NIRGraph.from_list(linear)
    module = nir_interpreter.nir_to_torch(graph, {})
    assert torch.allclose(module.linear.weight, torch.from_numpy(w))
    out = module(torch.ones(3))
    assert isinstance(out, typing.Tuple)
    assert out[0].shape == (2,)


def test_map_subgraph_default():
    w = np.random.random((2, 3)).astype(np.float32)
    linear = nir.Linear(w)
    subgraph = nir.NIRGraph.from_list(linear)
    graph = nir.NIRGraph.from_list(subgraph)
    module = nir_interpreter.nir_to_torch(graph, {})
    assert torch.allclose(module.nirgraph.linear.weight, torch.from_numpy(w))
    out = module(torch.ones(3))
    assert isinstance(out, typing.Tuple)
    assert out[0].shape == (2,)


def test_map_subgraph_with_state():
    tau = np.random.random(1)
    r = np.random.random(1)
    v_leak = np.random.random(1)
    li = nir.LI(tau, r, v_leak)
    li_module = _map_li(li)
    subgraph = nir.NIRGraph.from_list(li)
    graph = nir.NIRGraph.from_list(subgraph)
    module = nir_interpreter.nir_to_torch(graph, {nir.LI: _map_li})
    data = torch.ones(1)
    out = module(data)
    assert isinstance(out, typing.Tuple)
    assert out[0].shape == (1,)
    module_result = li_module(data)[0]
    assert torch.allclose(out[0], module_result)


def test_map_nested_subgraph_default():
    w = np.random.random((2, 3)).astype(np.float32)
    linear = nir.Linear(w)
    subgraph = nir.NIRGraph.from_list(linear)
    subgraph2 = nir.NIRGraph.from_list(subgraph)
    graph = nir.NIRGraph.from_list(subgraph2)
    module = nir_interpreter.nir_to_torch(graph, {})
    assert torch.allclose(module.nirgraph.nirgraph.linear.weight, torch.from_numpy(w))
    out = module(torch.ones(3))
    assert isinstance(out, typing.Tuple)
    assert out[0].shape == (2,)


def test_can_overwrite_default_map():
    w = np.random.random((2, 3)).astype(np.float32)
    linear = nir.Linear(w)
    graph = nir.NIRGraph.from_list(linear)
    state = {"called": False}

    def mock_linear_map(node: nir.Linear):
        state["called"] = True
        return torch.nn.Linear(2, 3)

    nir_interpreter.nir_to_torch(graph, {nir.Linear: mock_linear_map})
    assert state["called"]


def test_map_out_of_order():
    w = np.random.random((2, 3)).astype(np.float32)
    nodes = {
        "linear": nir.Linear(w),
        "input": nir.Input(np.array([2])),
        "output": nir.Output(np.array([3])),
    }
    edges = [("input", "linear"), ("linear", "output")]
    graph = nir.NIRGraph(nodes, edges)
    module = nir_interpreter.nir_to_torch(graph, {nir.Linear: _map_linear_node})
    data = torch.rand(3)
    assert torch.allclose(module(data)[0], torch.from_numpy(w) @ data)


# TODO: Implement recursive calls
@pytest.mark.skip("Recursion is not implemented yet")
def test_map_recursive_graph():
    w = np.random.random((2, 3)).astype(np.float32)
    nodes = {
        "linear": nir.Linear(w),
        "input": nir.Input(np.array([2])),
        "output": nir.Output(np.array([3])),
    }
    edges = [("linear", "linear"), ("input", "linear"), ("linear", "output")]
    graph = nir.NIRGraph(nodes, edges)
    module = nir_interpreter.nir_to_torch(graph, {nir.Linear: _map_linear_node})
    data = torch.rand(2)
    assert torch.allclose(module(data), data @ torch.from_numpy(w))


##########################################
#### Integration tests
##########################################


def test_from_nir_fx():
    w = np.random.random((1, 2))
    g = nir.NIRGraph.from_list(nir.Linear(w))
    with warnings.catch_warnings(record=True) as warn:
        m = nir_interpreter.nir_to_torch(g, _torch_node_map)
        assert len(warn) == 0

    assert m(torch.empty(2))[0].shape == (1,)


def test_import_lif_new_api():
    g = nir.read("tests/lif_norse.nir")
    m = nir_interpreter.nir_to_torch(g, _torch_node_map)
    assert m(torch.empty(1))[0].shape == (1,)



import warnings

import nir
import numpy as np
import pytest
import torch
import torch.nn as nn

from nirtorch.to_nir import extract_nir_graph


def _node_to_affine(node):
    if isinstance(node, torch.nn.Linear):
        return nir.Affine(node.weight.detach().numpy(), node.bias.detach().numpy())


def test_extract_single():
    m = nn.Linear(1, 1)
    g = extract_nir_graph(m, _node_to_affine, torch.rand(1, 1))
    assert set(g.edges) == {("input", "model"), ("model", "output")}
    assert isinstance(g.nodes["input"], nir.Input)
    assert np.allclose(g.nodes["input"].input_type["input"], np.array([1, 1]))
    assert isinstance(g.nodes["model"], nir.Affine)
    assert np.allclose(g.nodes["model"].weight, m.weight.detach().numpy())
    assert np.allclose(g.nodes["model"].bias, m.bias.detach().numpy())
    assert isinstance(g.nodes["output"], nir.Output)


def test_extract_nir_edges():
    mymodel = nn.Sequential(
        nn.Conv2d(2, 10, 3),
        nn.ReLU(),
        nn.AvgPool2d(2),
        nn.Conv2d(10, 4, 3),
        nn.ReLU(),
        nn.AvgPool2d(2),
    )
    sample_data = torch.rand(1, 2, 16, 16)

    def dummy_model_map(module: nn.Module) -> nir.NIRNode:
        return nir.NIRNode()

    nir_graph = extract_nir_graph(mymodel, dummy_model_map, sample_data, "mysequential")

    assert len(nir_graph.nodes) == 8
    print(nir_graph.edges)
    assert set(nir_graph.edges) == {
        ("input", "0"),
        ("0", "1"),
        ("1", "2"),
        ("2", "3"),
        ("3", "4"),
        ("4", "5"),
        ("5", "output"),
    }


class BranchedModel(nn.Module):
    def __init__(self, a, b, c):
        super().__init__()
        self.a = torch.randn(a, b)
        self.b = torch.randn(b, c)

    def forward(self, x):
        return x @ self.a @ self.b


@pytest.mark.skip(reason="Re-implement with correct recursive graph parsing")
def test_extract_multiple_explicit():
    model = nn.Sequential(BranchedModel(1, 2, 3), nn.Linear(3, 4))

    def extractor(module: nn.Module):
        if isinstance(module, BranchedModel):
            return nir.NIRGraph(
                nodes={
                    "0": nir.Input(np.array(module.a.shape[0])),
                    "1": nir.Linear(module.a),
                    "2": nir.Linear(module.b),
                    "3": nir.Output(np.array(module.b.shape[0])),
                },
                edges=[("0", "1"), ("0", "2"), ("1", "2"), ("1", "3"), ("2", "3")],
            )
        else:
            return nir.Affine(module.weight, module.bias)

    g = extract_nir_graph(model, extractor, torch.rand(1))
    print([type(n) for n in g.nodes])
    assert len(g.nodes) == 4
    assert len(g.edges) == 8  # in + 5 + 1 + out


def test_extract_recursive():
    class RecursiveModel(torch.nn.Module):
        def forward(self, x, s=None):
            if s is None:
                s = torch.zeros_like(x)
            return x + 1, s + x

    model = RecursiveModel()

    def extractor(m):
        if isinstance(m, RecursiveModel):
            return nir.Delay(np.array([1]))

    g = extract_nir_graph(model, extractor, torch.rand(1))
    assert set(g.edges) == {
        ("input", "model"),
        ("model", "output"),
        # ("model", "model") TODO: Detect and add recursive connections
    }


def test_ignore_batch_dim():
    model = nn.Linear(3, 1)

    def extractor(module: nn.Module):
        return nir.Affine(module.weight, module.bias)

    raw_input_shape = (1, 3)
    g = extract_nir_graph(
        model, extractor, torch.ones(raw_input_shape), ignore_dims=[0]
    )
    exp_input_shape = (3,)
    assert np.all(g.nodes["input"].input_type["input"] == np.array(exp_input_shape))
    assert g.nodes["model"].weight.shape == (1, 3)
    assert np.all(g.nodes["output"].output_type["output"] == np.array([1]))


def test_ignore_time_and_batch_dim():
    model = nn.Linear(3, 1)

    def extractor(module: nn.Module):
        return nir.Affine(module.weight, module.bias)

    raw_input_shape = (1, 10, 3)
    g = extract_nir_graph(
        model, extractor, torch.ones(raw_input_shape), ignore_dims=[0, -2]
    )
    exp_input_shape = (3,)
    assert np.all(g.nodes["input"].input_type["input"] == np.array(exp_input_shape))
    assert g.nodes["model"].weight.shape == (1, 3)

    raw_input_shape = (1, 10, 3)
    g = extract_nir_graph(
        model, extractor, torch.ones(raw_input_shape), ignore_dims=[0, 1]
    )
    exp_input_shape = (3,)
    assert np.all(g.nodes["input"].input_type["input"] == np.array(exp_input_shape))


# def test_extract_stateful():
#     model = norse.SequentialState(norse.LIFBoxCell(), nn.Linear(3, 1))

#     def extract(module: torch.nn.Module):
#         if isinstance(module, norse.LIFBoxCell):
#             return nir.NIR(
#                 nodes=[
#                     nir.LIF(
#                         tau=module.p.tau_mem_inv.detach(),
#                         v_threshold=module.p.v_th.detach(),
#                         v_leak=module.p.v_leak.detach(),
#                         r=torch.ones_like(module.p.v_leak.detach()),
#                     )
#                 ]
#             )
#         elif isinstance(module, torch.nn.Linear):
#             return nir.NIR(nodes=[nir.Linear(module.weight, module.bias)])

#     graph = extract_nir_graph(model, extract, torch.rand(1, 3))
#     assert len(graph.nodes) == 4
#     assert len(graph.edges) == 3
#     assert graph.edges[0] == (0, 1)


def test_deprecation_warning():
    with warnings.catch_warnings(record=True) as warn:
        m = nn.Linear(1, 1)
        extract_nir_graph(m, _node_to_affine, torch.rand(1, 1))
        assert len(warn) == 1
        assert isinstance(warn[0].message, DeprecationWarning)



from typing import Any, Callable, Dict, Set, Tuple
import operator

import numpy as np

import nir
import torch


DEFAULT_MAP: Dict[torch.nn.Module, Callable[[torch.nn.Module], nir.NIRNode]] = {
    torch.nn.Linear: (
        lambda module: nir.Affine(
            module.weight.detach().numpy(), module.bias.detach().numpy()
        )
    )
    # TODO: Add more default nodes
    # https://github.com/neuromorphs/NIRTorch/issues/25
}


class NIRTorchTracer(torch.fx.Tracer):

    def __init__(self, custom_leaf_modules: Tuple[torch.nn.Module] = None, **kwargs):
        """Extends PyTorch's default symbolic tracing with a set of custom leaf nodes"""
        super().__init__(**kwargs)
        if custom_leaf_modules is not None and not isinstance(
            custom_leaf_modules, tuple
        ):
            custom_leaf_modules = tuple(custom_leaf_modules)
        self.custom_leaf_modules = custom_leaf_modules

    def is_leaf_module(self, m: torch.nn.Module, module_qualified_name: str) -> bool:
        """
        A method to specify whether a given ``nn.Module`` is a "leaf" module.
        Leaf modules are the atomic units that appear in
        the IR, referenced by ``call_module`` calls. By default,
        Modules in the PyTorch standard library namespace (torch.nn)
        are leaf modules. All other modules are traced through and
        their constituent ops are recorded, unless specified otherwise
        via this parameter.
        Args:
            m (Module): The module being queried about
            module_qualified_name (str): The path to root of this module. For example,
                if you have a module hierarchy where submodule ``foo`` contains
                submodule ``bar``, which contains submodule ``baz``, that module will
                appear with the qualified name ``foo.bar.baz`` here.
        """
        # Tests that the module is in the list of custom leaves
        if self.custom_leaf_modules and isinstance(m, self.custom_leaf_modules):
            return True

        if hasattr(m, "_is_leaf_module") and m._is_leaf_module:
            return True

        return super().is_leaf_module(m, module_qualified_name)


class NIRTorchTransformer(torch.fx.Transformer):
    def call_function(self, target: str, args: Tuple, kwargs: Dict) -> Any:
        print("sup", target)
        return super().call_function(target, args, kwargs)

    def call_method(self, target: str, args: Tuple, kwargs: Dict) -> Any:
        return super().call_method(target, args, kwargs)

    def call_module(self, target, args, kwargs):
        print("mod", target)
        return super().call_module(target, args, kwargs)


def torch_to_nir(
    module: torch.nn.Module,
    module_map: Dict[torch.nn.Module, Callable[[torch.nn.Module], nir.NIRNode]],
    default_dict: Dict[
        torch.nn.Module, Callable[[torch.nn.Module], nir.NIRNode]
    ] = DEFAULT_MAP,
) -> nir.NIRGraph:
    """
    Traces a PyTorch module and converts it to a NIR graph using the specified module map.

    >>> import nir, nirtorch, numpy as np, torch
    >>> # First, we describe the PyTorch module we want to convert
    >>> torch_module = torch.nn.AvgPool2d(kernel_size=(2, 2), stride=0, padding=1)
    >>> # Second, we define the dictionary
    >>> torch_to_nir_map = {
    >>>     torch.nn.AvgPool2d: lambda module: nir.AvgPool2d(
    >>>         kernel_size=np.array(module.kernel_size),
    >>>         stride=np.array(module.stride),
    >>>         padding=np.array(module.padding)
    >>>     )
    >>> }
    >>> # Finally, we call nirtorch with the node and dictionary
    >>> torch_module = nirtorch.torch_to_nir(torch_module, torch_to_nir_map)
    >>> output, state = torch_module(input)        # Note the module returns a tuple of (output, state)
    >>> output, state = torch_module(input, state) # This can go on for many (time)steps

    Args:
        module (torch.nn.Module): The module of interest
        module_map (Dict[torch.nn.Module, Callable[[torch.nn.Module], nir.NIRNode]]): A dictionary that maps
            a given module type to a function that can convert the model to an NIRNode type
        default_dict (Dict[torch.nn.Module, Callable[[torch.nn.Module], nir.NIRNode]]): An dictionary
            of default mappings that, by default, maps trivial modules like torch.nn.Linear. Override
            the dictionary to provide custom mappings.
    """
    # Merge the default dictionary, if it exists
    if default_dict is not None:
        module_map = module_map | default_dict

    # Cover the edge case that the incoming module is a leaf node
    if module.__class__ in module_map:
        return module_map[module.__class__](module)

    # Trace the graph
    tracer = NIRTorchTracer(module_map.keys())
    traced = tracer.trace(module)

    if len(traced.nodes) == 2 and len(list(tracer.root.children())) == 0:
        raise ValueError(
            "The module is a leaf node, but does not appear in the module map. We cannot trace it further"
        )

    graph_module = torch.fx.GraphModule(tracer.root, traced)

    # Create NIR nodes
    nodes = {}
    edges = []
    ignored_nodes = set()
    bypass_nodes = set()

    def _find_users(node: torch.fx.Node) -> Set[torch.fx.Node]:
        """
        Finds all the users (outputs) of a given node, recursively if the node is registered as a bypass node
        """
        nodes = set()
        for user in node.users:
            if user in ignored_nodes:
                continue
            elif user in bypass_nodes:
                nodes |= _find_users(user)
            else:
                nodes.add(user)
        return nodes

    def _find_inputs(node: torch.fx.Node) -> Set[torch.fx.Node]:
        """
        Finds all the inputs (inputs) of a given node, recursively if the node is registered as a circumvented node
        """
        nodes = set()
        for in_node in node.all_input_nodes:
            if in_node in ignored_nodes:
                continue
            elif in_node in bypass_nodes:
                nodes |= _find_inputs(in_node)
            else:
                nodes.add(in_node)
        return nodes

    for node in traced.nodes:
        # Add Node
        if node.op == "placeholder":
            if node.target == "input" or node.prev.op == "root":
                nodes[str(node.name)] = nir.Input(np.array([1]))
            else:
                ignored_nodes.add(node)
                continue
        elif node.op == "output":
            nodes[str(node.name)] = nir.Output(np.array([1]))
        elif node.op == "call_function":
            # Ensure that we bypass add nodes
            # TODO: Consider using transformations for this
            #       https://pytorch.org/docs/stable/fx.html#torch.fx.Transformer
            if node.target == operator.add:
                bypass_nodes.add(node)
            # Raise a warning if we encounter other methods than addition
            else:
                raise ValueError(
                    "The only supported function is addition. Please modify your model or raise an issue on GitHub"
                )
        elif node.op == "call_method":
            # Bypass add methods
            if node.target == "add":
                bypass_nodes.add(node)
            else:
                raise ValueError(
                    "The only supported method is addition. Please modify your model or raise an issue on GitHub"
                )
        elif node.op == "call_module":
            torch_module = graph_module.get_submodule(node.target)
            nir_module = module_map[torch_module.__class__](torch_module)
            nodes[str(node.name)] = nir_module
        elif node.op == "get_attr":
            # Bypass attribute
            bypass_nodes.add(node)
        else:
            raise ValueError(
                f"Unsupported operation {node.op}. Please modify your model or raise an issue on GitHub"
            )

    # Create edges
    # - This is done in a separate loop to ensure that we correctly ignore the edges in case the nodes
    #   are ignored out-of-order
    for node in traced.nodes:
        if node in ignored_nodes:
            continue

        # Add edges
        for in_node in node.all_input_nodes:
            if in_node in ignored_nodes or in_node in bypass_nodes:
                continue
            # If the function is set to be bypassed, we simply forward the input to all the outputs
            if node in bypass_nodes:
                for next_node in _find_users(node):
                    edges.append((in_node.name, next_node.name))
            # Ignore additions as incoming edges
            elif in_node.op == "call_function" and in_node.target == operator.add:
                break
            # Otherwise, add an edge
            elif in_node not in ignored_nodes:
                edges.append((in_node.name, node.name))
    graph = nir.NIRGraph(nodes=nodes, edges=edges)
    graph.infer_types()
    return graph


if __name__ == "__main__":
    module = torch.nn.Sequential(torch.nn.Linear(2, 1), torch.nn.Linear(1, 1))
    graph = torch_to_nir(module)

    import pprint

    pprint.pprint(graph)
